#ifndef __RESULT_HPP__
#define __RESULT_HPP__


#include <exception>
#include <memory>
#include <type_traits>


namespace typon
{

  template <typename T>
  struct result
  {
    std::exception_ptr _exception;
    union
    {
      T _value;
    };

    ~result()
    {
      if (!_exception)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    template <typename U>
    void return_value(U&& expr) noexcept(std::is_nothrow_constructible_v<T, U&&>)
    {
      std::construct_at(std::addressof(_value), std::forward<U>(expr));
    }

    void unhandled_exception() noexcept
    {
      _exception = std::current_exception();
    }

    T& get() &
    {
      if (_exception)
      {
        std::rethrow_exception(_exception);
      }
      return _value;
    }

    T&& get() &&
    {
      if (_exception)
      {
        std::rethrow_exception(_exception);
      }
      return std::move(_value);
    }
  };


  template <typename T>
  struct result<T&>
  {
    T* _value;
    std::exception_ptr _exception;

    void return_value(T& expr) noexcept
    {
      _value = std::addressof(expr);
    }

    void unhandled_exception() noexcept
    {
      _exception = std::current_exception();
    }

    T& get() &
    {
      if (_exception)
      {
        std::rethrow_exception(_exception);
      }
      return *_value;
    }
  };


  template <>
  struct result<void>
  {
    std::exception_ptr _exception;

    void return_void() noexcept {}

    void unhandled_exception() noexcept
    {
      _exception = std::current_exception();
    }

    void get()
    {
      if (_exception)
      {
        std::rethrow_exception(_exception);
      }
    }
  };

}

#endif // __RESULT_HPP__
