#ifndef __TASK_HPP__
#define __TASK_HPP__

#include <coroutine.hpp>
#include <result.hpp>


namespace typon
{

  template <typename T = void>
  struct [[nodiscard]] task
  {
    struct promise_type;

    coroutine_handle<promise_type> _coroutine;

    ~task()
    {
      _coroutine.destroy();
    }

    struct promise_type : result<T>
    {
      coroutine_handle<> _continuation;

      task get_return_object() noexcept
      {
        return { coroutine_handle<promise_type>::from_promise(*this) };
      }

      suspend_never initial_suspend() noexcept
      {
        return {};
      }

      auto final_suspend() noexcept
      {
        struct awaitable : suspend_always
        {
          coroutine_handle<> await_suspend(coroutine_handle<promise_type> coroutine) noexcept
          {
            if (coroutine.promise()._continuation)
            {
              return coroutine.promise()._continuation;
            }
            return noop_coroutine();
          }
        };

        return awaitable {};
      }
    };

    auto operator co_await() && noexcept
    {
      struct awaitable
      {
        coroutine_handle<promise_type> _coroutine;

        bool await_ready() noexcept
        {
          return _coroutine.done();
        }

        coroutine_handle<> await_suspend(coroutine_handle<> awaiting_coroutine) noexcept
        {
          _coroutine.promise()._continuation = awaiting_coroutine;
          return _coroutine;
        }

        decltype(auto) await_resume()
        {
          return _coroutine.promise().get();
        }
      };

      return awaitable { _coroutine };
    }

    decltype(auto) call() &&
    {
      return _coroutine.promise().get();
    }
  };
  
}

#endif // __TASK_HPP__
